#define sei SREG_I_bit = 1               // Interrupt enable
#define cli SREG_I_bit = 0               // Interrupt disable
/* os definition */
#define t_idel         0x00
#define t_activate     0x01
#define t_run          0x02
#define t_interrupt    0x04
#define t_suspend      0x08


/* osc = 8000000 8MHz
   prescales is set to 1/256  */
#define prescaler 0x06

/* interupt is generated for 2.5 ms tick */
#define CCOVF 0x4E

#include "taskdefs.h"
/* tick count for scheduler */
volatile unsigned char tick;

/* background task status flag */
volatile unsigned char b_status;

/* kernel address to be call form timer isr */
unsigned int kernel_address;
volatile unsigned char kh;
volatile unsigned char kl;

/* task information */
typedef struct {
unsigned char taskstatus;
unsigned char tickcount;
unsigned char decrement;
} task_info;

/* isr is called every 2.5 ms at 8Mhz osc */
void Timer0_COMP_Interrupt() iv IVT_ADDR_TIMER0_COMP ics ICS_OFF
{
/* -----------------------------------------------------
This function is a tick generator and also the kernel execution.
is save the current content and then call the scheduler for next task execution
in case of background task the content is save and the current task is started.
after the execution of all tasks the background task is resumed.
----------------------------------------------------- */
        /* save status register */
        asm { push r16
              in r16,SREG
              push r16        };
        /* increment tick counter */
        asm {    lds r16,_tick
                 inc r16
                 brcc 1f
                 inc r16
                 1f:
                 sts _tick,r16};

        /* check if the background task is active */
        asm {lds r16,_b_status          ; load background task flag
                 subi r16,02            ; check if background task is running
                 BREQ 2f                ; go to save context and call kernel
                 pop r16                ; no background task activate
                 out SREG,r16           ; resore register
                 pop r16
                 reti                   ; return to execution
                 2f:
                  ldi r16,0x04          ; set background task statu to suspend
                 sts _b_status,r16
                 push R0                ; save all regiter on stack for future
                 push R1                ; restore
                 push R2
                 push R3
                 push R4
                 push R5
                 push R6
                 push R7
                 push R8
                 push R9
                 push R10
                 push R11
                 push R12
                 push R13
                 push R14
                 push R15
                 push R17
                 push R18
                 push R19
                 push R20
                 push R21
                 push R22
                 push R23
                 push R24
                 push R25
                 push R26
                 push R27
                 push R28
                 push R29
                 push R30
                 push R31
                 lds r16,_kl           ; load low address of kernel
                 push r16
                 lds r16,_kh           ; load high address of kernel
                 push r16
                 ldi r16,0x00          ; force SREG to 0
                 out SREG,r16
                 reti                  ; force return to kernel
                };
}

void os_startup (void)
{
 // setup prescaler
        OCR0 = CCOVF;
        TCCR0 = prescaler | 0x08;
        TCNT0 = 0x00;
        TIMSK = 0x02;
/* power mode idle */
        MCUCR = 0x20;
        tick = 0;
        sei;
}

/* nomber of task to define */
#define tasks_no 3

task_info os_task[tasks_no];

/* pointer with all task in list */
void (*f[])(void) = { f1, f2, f3, backgroundtask };

void kernel ( void );

/* setup all tasks - time base and status idle */
void task_setup(void)
{
        os_task[0].taskstatus = t_idel;
        os_task[0].tickcount = 2;
        os_task[0].decrement = os_task[0].tickcount;

        os_task[1].taskstatus = t_idel;
        os_task[1].tickcount = 4;
        os_task[1].decrement = os_task[1].tickcount;

        os_task[2].taskstatus = t_idel;
        os_task[2].tickcount = 40;
        os_task[2].decrement = os_task[2].tickcount;

        b_status = t_idel;

        kernel_address = (unsigned int ) &kernel;
        kh = (char) (kernel_address >> 8);
        kl = (char) (kernel_address );

}


void kernel ( void )
{
unsigned char i;

 while (1)
 {
       if (tick != 0)
       {
               /* Decrement all task tick */
         cli;
                 TIMSK = TIMSK & 0xFD;
         sei;
         for (i= 0; i <(tasks_no) ; i++)
         {
                 /* !! suspend the timer interupt durring this activities !! */
                /* check is the task is suspendet */
                if (os_task[i].taskstatus != t_suspend)
                {
                /* for all not suspende task decrement tick */
                        os_task[i].decrement = os_task[i].decrement - tick;
                        /* if we lost some tick recover and activate the request task */
                        if (os_task[i].decrement > os_task[i].tickcount)
                        {
                                os_task[i].decrement = 255 - os_task[i].decrement ;
                                os_task[i].decrement = os_task[i].tickcount - os_task[i].decrement -1;
                                /* if the task is running do not activate again. the task will lost a period
                                   only the idle task will be activated                                                                          */
                                if(os_task[i].taskstatus == t_idel)
                                {
                                        /* task is set to run */

                                                os_task[i].taskstatus = t_activate;

                                }
                                else
                                {
                                  /* nothing */
                                }
                        }
                        else
                        {
                                /* nothing */
                        }
                }
                else
                {
                }
         }

          tick=0;
        /* !! resume here timer interupt !! */
         cli;
                 TIMSK = TIMSK | 0x02;
         sei;
         /* activate all the request */
         for (i=0 ;i<(tasks_no);i++)
         {
                if (os_task[i].taskstatus == t_activate)
                {
                        os_task[i].taskstatus = t_run;
                        f[i]();
                        cli;
                        os_task[i].taskstatus = t_idel;
                        sei;
                }

          }
/* activate the background task
 before activating the background task check if we have some other task to
activat in case of jiter */
        if (tick==0)
        {
         cli;
         if (b_status == t_idel)
         {
            b_status = t_run;
            f[tasks_no]();
            cli;
            b_status = t_idel;
            sei;
         }
         else if (b_status == t_interrupt)
         {
              b_status = t_run;
              asm {
                   pop r31
                   pop R30
                   pop R29
                   pop R28
                   pop R27
                   pop R26
                   pop R25
                   pop R24
                   pop R23
                   pop R22
                   pop R21
                   pop R20
                   pop R19
                   pop R18
                   pop R17
                   pop R15
                   pop R14
                   pop R13
                   pop R12
                   pop R11
                   pop R10
                   pop R9
                   pop R8
                   pop R7
                   pop R6
                   pop R5
                   pop R4
                   pop R3
                   pop R2
                   pop R1
                   pop R0
                   pop r16
                   out SREG,r16
                   pop r16
                   reti
                  }
         }
        }
        
       }
       
        if(tick==0)
        {
                asm { sleep };
        }

 }
}

void main() 
{
 task_setup();
 os_startup();
 kernel();
/* code never executed . just guard */
 while(1);
}